/*
 *  IHES.Agent network command interface
 *  By Shikai Chen
 *
 */

#pragma once

namespace ihes{namespace net{

#ifndef _WIN32
#include <netinet/tcp.h>

#else

#endif

#define CONNECTION_QLEN 10


//agent side
class CmdHandler
{
public:
    CmdHandler(u16 agent_type)
        : _socket_fd(0)
        , _agentType(agent_type)
    {
        init_socket();
        bind();
    }

    ~CmdHandler()
    {
        release_socket();
    }

    bool bind()
    {
        if ( _socket_fd == 0 ) return false;

        
        bzero( &_server_addr, sizeof(_server_addr) );
        _server_addr.sin_family = AF_INET;

        //TODO: maybe changed
        _server_addr.sin_addr.s_addr = htonl(INADDR_ANY); //listen to all network
        _server_addr.sin_port = htons(AGENT_PORT(_agentType));

        ::bind(_socket_fd, (struct sockaddr *)&_server_addr, (socklen_t)sizeof(_server_addr));
        return true;
    }

    u_result listenNext(sockaddr * addr = NULL, socklen_t *addrlen= NULL)
    {
        if (listen(_socket_fd, CONNECTION_QLEN) < 0) {
            return -1;
        }

        int     destfd;

        destfd = accept(_socket_fd, addr, addrlen);
        
        if (destfd < 0) {
            return -1;
        }

#ifndef _WIN32
        timeval tv;
        tv.tv_sec = 10; 
        tv.tv_usec = 0; 

        setsockopt( destfd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv) );
        setsockopt( destfd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv) );
#else
        
#endif
        int bool_true = 1;
        setsockopt( destfd, IPPROTO_TCP, TCP_NODELAY, (char *)&bool_true, sizeof(bool_true) );
        return destfd;
    }

    u_result recv_cmd(int  destfd, u16 &maincmd, u16 &subcmd)
    {
        unsigned char databuf[sizeof(PKT_AGENT_CMD)];

        int received_len = 0;

        int current_recved ;
        while ( (current_recved = recv( destfd, (char *)&databuf[received_len], sizeof(PKT_AGENT_CMD) - received_len, 0 )) >0 )
        {   
            received_len += current_recved;

            if ( received_len>=sizeof(PKT_AGENT_CMD)) break;
        }

        if (received_len<sizeof(PKT_AGENT_CMD))   return -1;

        PKT_AGENT_CMD *pkt = (PKT_AGENT_CMD *)databuf;

        if ( pkt->header.isValid() && pkt->header.pkt_type == PKT_TYPE_CMD )
        {
            maincmd = pkt->cmd;
            subcmd = pkt->subcmd;
            return pkt->getDataSize();
        }
        return -1;
    }

    u_result recv_remaining(int  destfd, unsigned char * data, u32 datalen)
    {
        if (data == NULL || datalen==0) return 0;

        int received_len = 0;
        int current_recved ;

        while ( (current_recved = recv( destfd, (char *)&data[received_len], datalen - received_len, 0 )) >0 )
        {   
            received_len += current_recved;

            if ( received_len>=datalen) break;
        }

        return received_len;
    }

    u_result send_ans(int  destfd, u16 anscmd, u32 ans = 0, unsigned char * data = NULL, u32 datalen = 0)
    {
        int sent_len = 0;
        int current_sent ;
        int pkt_size;

        if (data == NULL) datalen =0 ;

        pkt_size = sizeof(PKT_AGENT_ANS) + datalen;

        unsigned char * buffer = new unsigned char[pkt_size];
        if (buffer == 0) return -1;

        PKT_AGENT_ANS * ansPkt = new (buffer) PKT_AGENT_ANS(datalen, _agentType, anscmd, ans);
        if (datalen>0)
            memcpy( &buffer[sizeof(PKT_AGENT_ANS)], data, datalen);

        while ( ( current_sent = send( destfd, (char *)&buffer[sent_len], pkt_size - sent_len, 0 )) > 0 )
        {
            sent_len += current_sent;

            if ( sent_len>=pkt_size) break;
        }
        

        delete [] buffer;
        return sent_len;
    }

protected:
    bool    init_socket()
    {
        //TODO: assuming always uses TCP, may be changed later
        _socket_fd = socket(AF_INET, SOCK_STREAM, 0);
        if ( _socket_fd == -1 ) 
        {
            _socket_fd = 0;
            return false;
        }

        int bool_true = 1;

        setsockopt( _socket_fd, SOL_SOCKET, SO_REUSEADDR , (char *)&bool_true, sizeof(bool_true) );

        
        //TODO: timeout value should be set
        return true;
    }   


    void    release_socket()
    {
#ifndef _WIN32
        ::close(_socket_fd);
#else
        closesocket(_socket_fd);
#endif
    }

    u16                 _agentType;
    int                _socket_fd;

    struct sockaddr_in _server_addr;
};


//server side
class CmdSender
{
public:
    CmdSender()
        : _socket_fd(0)
        , _is_agent_connected(false)
        , _dest_agentType(0)
    {
        //init_socket();

    }

    ~CmdSender()
    {
        release_socket();
    }

    u_result connectAgent(u16 agentType, const char * addr = NULL)
    {
        if (_is_agent_connected)
        {
            close();
        }

        if (addr == NULL) addr = "localhost";
        hostent *dest_add = gethostbyname(addr);
        if (dest_add == NULL)
        {
            return -1;
        }

        bzero(&_agent_addr,sizeof(_agent_addr));

        _agent_addr.sin_family = AF_INET;
        _agent_addr.sin_port = htons(AGENT_PORT(agentType));
        _agent_addr.sin_addr = *((in_addr *)dest_add->h_addr);
        
        if (!init_socket()) return -1;

        if (connect(_socket_fd, (sockaddr *)&_agent_addr, sizeof(_agent_addr)) != 0) 
        {
            release_socket();
            return -1;
        }

        _dest_agentType = agentType;
        _is_agent_connected = true;

        return 0;
    }

    void close()
    {
        release_socket();
        _is_agent_connected = true;
    }


    u_result recv_ans(u16 &anscmd, u32 &result)
    {
        if (!_is_agent_connected) return -1;

        unsigned char databuf[sizeof(PKT_AGENT_ANS)];

        int received_len = 0;

        int current_recved ;
        while ( (current_recved = recv( _socket_fd, (char *)&databuf[received_len], sizeof(databuf) - received_len, 0 )) >0 )
        {   
            received_len += current_recved;

            if ( received_len>=sizeof(databuf)) break;
        }

        if (received_len<sizeof(databuf))   return -1;

        PKT_AGENT_ANS *pkt = (PKT_AGENT_ANS *)databuf;

        if ( pkt->header.isValid() && pkt->header.pkt_type == PKT_TYPE_ANS && pkt->agent_type == _dest_agentType)
        {
            anscmd = pkt->cmd;
            result = pkt->result;
            return pkt->getDataSize();
        }
        return -1;
    }

    u_result recv_remaining( unsigned char * data, u32 datalen)
    {
        if (data == NULL || datalen==0) return 0;

        int received_len = 0;
        int current_recved ;

        while ( (current_recved = recv( _socket_fd, (char *)&data[received_len], datalen - received_len, 0 )) >0 )
        {   
            received_len += current_recved;

            if ( received_len>=datalen) break;
        }

        return received_len;
    }

    u_result send_cmd(u16 maincmd, u16 subcmd = 0, unsigned char * data = NULL, u32 datalen = 0)
    {
        int sent_len = 0;
        int current_sent ;
        int pkt_size;

        if (data == NULL) datalen =0 ;

        pkt_size = sizeof(PKT_AGENT_CMD) + datalen;

        unsigned char * buffer = new unsigned char[pkt_size];
        if (buffer == 0) return -1;

        PKT_AGENT_CMD * ansPkt = new (buffer) PKT_AGENT_CMD(datalen,  maincmd, subcmd);
        if (datalen>0)
            memcpy( &buffer[sizeof(PKT_AGENT_CMD)], data, datalen);

        while ( ( current_sent = send( _socket_fd, (char *)&buffer[sent_len], pkt_size - sent_len, 0 )) > 0 )
        {
            sent_len += current_sent;

            if ( sent_len>=pkt_size) break;
        }
        

        delete [] buffer;
        return sent_len;
    }

protected:
    bool    init_socket()
    {
        //TODO: assuming always uses TCP, may be changed later
        _socket_fd = socket(AF_INET, SOCK_STREAM, 0);
        if ( _socket_fd == -1 ) 
        {
            _socket_fd = 0;
            return false;
        }

        int bool_true = 1;

        setsockopt( _socket_fd, SOL_SOCKET, SO_REUSEADDR , (char *)&bool_true, sizeof(bool_true) );
        setsockopt( _socket_fd, IPPROTO_TCP, TCP_NODELAY,(char *) &bool_true, sizeof(bool_true) );

#ifndef _WIN32
        timeval tv;
        tv.tv_sec = 10; 
        tv.tv_usec = 0; 

        setsockopt( _socket_fd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv) );
        setsockopt( _socket_fd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv) );
#else

#endif
        return true;
    }   


    void    release_socket()
    {
#ifndef _WIN32
        ::close(_socket_fd);
#else
        closesocket(_socket_fd);
#endif
    }

    bool               _is_agent_connected;
    int                _socket_fd;
    struct sockaddr_in _agent_addr;

    u16                _dest_agentType;
};

}}
